@@ -3,8 +3,7 @@
 /*               This file is part of the program and library                */
 /*    PaPILO --- Parallel Presolve for Integer and Linear Optimization       */
 /*                                                                           */
-/* Copyright (C) 2020-2023 Konrad-Zuse-Zentrum                               */
-/*                     fuer Informationstechnik Berlin                       */
+/* Copyright (C) 2020-2024 Zuse Institute Berlin (ZIB)                       */
 /*                                                                           */
 /* This program is free software: you can redistribute it and/or modify      */
 /* it under the terms of the GNU Lesser General Public License as published  */
@@ -621,23 +620,13 @@ update_activity_after_coeffchange( REAL collb, REAL colub, ColFlags cflags,
 /// and is called to inform about column bounds that changed.
 template <typename REAL, typename BOUNDCHANGE>
 void
-propagate_row( int row, const REAL* rowvals, const int* colindices, int rowlen,
+propagate_row( const Num<REAL>& num, int row, const REAL* rowvals, const int* colindices, int rowlen,
                const RowActivity<REAL>& activity, REAL lhs, REAL rhs,
                RowFlags rflags, const Vec<REAL>& lower_bounds,
                const Vec<REAL>& upper_bounds, const Vec<ColFlags>& domainFlags,
                BOUNDCHANGE&& boundchange )
 {
-
-   bool adj_rhs = false;
-   if( activity.ninfmin == 1 && activity.ninfmax == 0 &&
-       rflags.test( RowFlag::kRhsInf ) )
-   {
-      adj_rhs = true;
-      rhs = activity.max;
-   }
-
-   if( ( !rflags.test( RowFlag::kRhsInf ) && activity.ninfmin <= 1 ) ||
-       adj_rhs )
+   if( !rflags.test( RowFlag::kRhsInf ) && activity.ninfmin <= 1 && ( activity.ninfmax >= 1 || num.isGT(activity.max, rhs) ) )
    {
       for( int j = 0; j < rowlen; ++j )
       {
@@ -647,7 +636,7 @@ propagate_row( int row, const REAL* rowvals, const int* colindices, int rowlen,
          REAL minresact = activity.min;
          REAL val = rowvals[j];
 
-         if( val < REAL{ 0.0 } )
+         if( val < 0 )
          {
             if( activity.ninfmin == 1 )
             {
@@ -661,10 +650,15 @@ propagate_row( int row, const REAL* rowvals, const int* colindices, int rowlen,
                assert( !domainFlags[col].test( ColFlag::kUbUseless ) );
                minresact -= val * ub;
             }
-
-            REAL newlb = ( rhs - minresact ) / val;
-            if( domainFlags[col].test( ColFlag::kLbInf ) || newlb > lb )
-               boundchange( BoundChange::kLower, col, newlb, row );
+            REAL new_lb = (rhs - minresact) / val;
+            if( domainFlags[col].test( ColFlag::kIntegral ) )
+            {
+               new_lb = floor(new_lb);
+               if( num.isFeasLT(rhs, new_lb * val + minresact) )
+                  new_lb += 1;
+            }
+            if( domainFlags[col].test( ColFlag::kLbInf ) || new_lb > lb )
+               boundchange( BoundChange::kLower, col, new_lb, row );
          }
          else
          {
@@ -680,24 +674,20 @@ propagate_row( int row, const REAL* rowvals, const int* colindices, int rowlen,
                assert( !domainFlags[col].test( ColFlag::kLbUseless ) );
                minresact -= val * lb;
             }
-
-            REAL newub = ( rhs - minresact ) / val;
-            if( domainFlags[col].test( ColFlag::kUbInf ) || newub < ub )
-               boundchange( BoundChange::kUpper, col, newub, row );
+            REAL new_ub = (rhs - minresact) / val;
+            if( domainFlags[col].test( ColFlag::kIntegral ) )
+            {
+               new_ub = ceil(new_ub);
+               if( num.isFeasLT(rhs, new_ub * val + minresact) )
+                  new_ub -= 1;
+            }
+            if( domainFlags[col].test( ColFlag::kUbInf ) || new_ub < ub )
+               boundchange( BoundChange::kUpper, col, new_ub, row );
          }
       }
    }
 
-   bool adj_lhs = false;
-   if( activity.ninfmax == 1 && activity.ninfmin == 0 &&
-       rflags.test( RowFlag::kLhsInf ) )
-   {
-      adj_lhs = true;
-      lhs = activity.min;
-   }
-
-   if( ( !rflags.test( RowFlag::kLhsInf ) && activity.ninfmax <= 1 ) ||
-       adj_lhs )
+   if( !rflags.test( RowFlag::kLhsInf ) && activity.ninfmax <= 1 && ( activity.ninfmin >= 1 || num.isLT(activity.min, lhs) ) )
    {
       for( int j = 0; j < rowlen; ++j )
       {
@@ -707,7 +697,7 @@ propagate_row( int row, const REAL* rowvals, const int* colindices, int rowlen,
          REAL maxresact = activity.max;
          REAL val = rowvals[j];
 
-         if( val < REAL{ 0.0 } )
+         if( val < 0 )
          {
             if( activity.ninfmax == 1 )
             {
@@ -721,10 +711,15 @@ propagate_row( int row, const REAL* rowvals, const int* colindices, int rowlen,
                assert( !domainFlags[col].test( ColFlag::kLbUseless ) );
                maxresact -= val * lb;
             }
-
-            REAL newub = ( lhs - maxresact ) / val;
-            if( domainFlags[col].test( ColFlag::kUbInf ) || newub < ub )
-               boundchange( BoundChange::kUpper, col, newub, row );
+            REAL new_ub = (lhs - maxresact) / val;
+            if( domainFlags[col].test( ColFlag::kIntegral ) )
+            {
+               new_ub = ceil(new_ub);
+               if( num.isFeasGT(lhs, new_ub * val + maxresact) )
+                  new_ub -= 1;
+            }
+            if( domainFlags[col].test( ColFlag::kUbInf ) || new_ub < ub )
+               boundchange( BoundChange::kUpper, col, new_ub, row );
          }
          else
          {
@@ -740,10 +735,15 @@ propagate_row( int row, const REAL* rowvals, const int* colindices, int rowlen,
                assert( !domainFlags[col].test( ColFlag::kUbUseless ) );
                maxresact -= val * ub;
             }
-
-            REAL newlb = ( lhs - maxresact ) / val;
-            if( domainFlags[col].test( ColFlag::kLbInf ) || newlb > lb )
-               boundchange( BoundChange::kLower, col, newlb, row );
+            REAL new_lb = (lhs - maxresact) / val;
+            if( domainFlags[col].test( ColFlag::kIntegral ) )
+            {
+               new_lb = floor(new_lb);
+               if( num.isFeasGT(lhs, new_lb * val + maxresact) )
+                  new_lb += 1;
+            }
+            if( domainFlags[col].test( ColFlag::kLbInf ) || new_lb > lb )
+               boundchange( BoundChange::kLower, col, new_lb, row );
          }
       }
    }
